#include <LiquidCrystal_I2C.h>
#include <string.h>

#define ENDSTOP (5)
#define UP (4)
#define DOWN (3)
#define OK (2)
#define STEP (9)
#define DIR (10)

#define FORWARD LOW
#define REVERSE HIGH

#define HOMING_DELAY_US (20)

#define AXIS_LENGTH_MM (200)
#define STEPS_PER_MM (100)

// this is 20000 (20k)
#define AXIS_LENGTH (AXIS_LENGTH_MM * STEPS_PER_MM)

#define TO_STEPS(mms) (mms * STEPS_PER_MM)

char lines[4][17];

class EdgeDetector {
private:
  unsigned long last = 0;
  bool state = false;
public:
  bool feed(bool st) {
    unsigned long tnow = millis();       
    if (st && !state && (tnow - last) > 80) {
      last = tnow;
      state = st;
      return true;
    }
    state = st;
    return false;
  }
};

EdgeDetector butUp = EdgeDetector();
EdgeDetector butDown = EdgeDetector();
EdgeDetector butSelect = EdgeDetector();

/* Display layout:
 *
 * 123456789abcdef
 * * Points     10 | 2-10 points
 * * Delay      20 | 1-20 minutes
 * * Skip          |
 * 
 */

struct _disp {

  int focusLine = 0;
  int points = 2;
  int minutes = 1;
  bool focus = false;

  unsigned long timeLast = 0;


  void init() {
    timeLast = millis();
  }
  bool update() {
    int refreshDisp = false;
    unsigned long tnow  = millis();
    if (tnow - timeLast > 200) {
      timeLast = millis();
      refreshDisp = true;
    }

    return refreshDisp;
  }

} disp;

struct _path {
  int position = 0; // in steps
  int target = 0; // in mm
  bool pulse = LOW;
  bool enabled = false;
} path;

struct _planner {
private:
  int points = 2;
  int currentPoint = 0;
  int minutes = 1;
  
  unsigned long timeLast = 0;

  void setTarget(int pt) {
    path.target = AXIS_LENGTH_MM * ((pt * 1.0) / (points - 1));
    currentPoint = pt;
  }
public:
  void init() {
    timeLast = millis();
  }

  void setMinutes(int mn) {
    minutes = mn;
    timeLast = millis();
  }

  void setPoints(int pt) {
    points = pt;
    if (currentPoint >= pt) {
      setTarget(pt - 1);
    } else {
      setTarget(currentPoint);
    }
  }

  int getCurrentPoint() {
    return currentPoint;
  }

  void run() {
    unsigned long tnow  = millis();
    if (tnow - timeLast > 60000l * minutes) {
      timeLast = millis();
      setTarget(currentPoint == points - 1 ? 0 : currentPoint + 1);
    }
  }

  void next() {
    timeLast = millis();
    setTarget(currentPoint == points - 1 ? 0 : currentPoint + 1);
  }
} planner;

LiquidCrystal_I2C lcd(0x38, 16, 4);

void homeMachine() {
  for (int i = 1; i <= 5; i++) {
    int del = HOMING_DELAY_US * (i == 1 ? 1 : 40 * i);
    digitalWrite(DIR, REVERSE);
    while (digitalRead(ENDSTOP)) {
      digitalWrite(STEP, LOW);
      delayMicroseconds(del);
      digitalWrite(STEP, HIGH);
      delayMicroseconds(del);
    }
    digitalWrite(DIR, FORWARD);
    while (!digitalRead(ENDSTOP)) {
      digitalWrite(STEP, LOW);
      delayMicroseconds(del);
      digitalWrite(STEP, HIGH);
      delayMicroseconds(del);
    }
  }
  digitalWrite(STEP, LOW);
  digitalWrite(DIR, FORWARD);
  path.enabled = true;
}

ISR(TIMER2_COMPA_vect){
   if (!path.enabled) {
     return;
   }
   int target = TO_STEPS(path.target);
   if (path.position == target) {
     return;
   }
   bool dir = (path.position < target ? FORWARD : REVERSE);
   path.pulse = !path.pulse;
   digitalWrite(DIR, dir);
   digitalWrite(STEP, path.pulse);
   if (dir == FORWARD && path.pulse) {
     path.position++;
   }
   if (dir == REVERSE && path.pulse) {
     path.position--;
   }
}

void updateLcd() {
  sprintf(lines[0],  "  Current %-2d/%d ", planner.getCurrentPoint() + 1, disp.points);
  if (strlen(lines[0]) < 16) strcat(lines[0], " ");

  sprintf(lines[1], "%c %s%-2d", (disp.focusLine == 0 ? (disp.focus ? '*' : '>') : ' '), "Points     ", disp.points);
  sprintf(lines[2], "%c %s%-2d", (disp.focusLine == 1 ? (disp.focus ? '*' : '>') : ' '), "Minutes    ", disp.minutes);
  sprintf(lines[3], "%c Skip          ", (disp.focusLine == 2 ? '>' : ' '));
  //lcd.clear();
  Serial.println(strlen(lines[1]));
  lcd.setCursor(0, 0);
  lcd.print(lines[0]);
  lcd.setCursor(0, 1);
  lcd.print(lines[1]);
  lcd.setCursor(16, 0);
  lcd.print(lines[2]);
  lcd.setCursor(16, 1);
  lcd.print(lines[3]);
}

void handleUi() {
  bool sigDown = butDown.feed(digitalRead(DOWN));
  bool sigUp = butUp.feed(digitalRead(UP));
  bool sigSelect = butSelect.feed(digitalRead(OK));

  if (sigSelect) {
    if (disp.focusLine == 2) {
      planner.next();
    } else disp.focus = !disp.focus;
  }

  int modif = (sigDown ? -1 : 0) + (sigUp ? 1 : 0);
  if (disp.focus) {
    if (disp.focusLine == 0) {
      disp.points = constrain(disp.points + modif, 2, 20);
      planner.setPoints(disp.points);
    }
    if (disp.focusLine == 1) {
      disp.minutes = constrain(disp.minutes + modif, 1, 20);
      planner.setMinutes(disp.minutes);
    }    
  } else {
    int nextline = disp.focusLine - modif;
    if (nextline == -1) nextline = 2;
    if (nextline == 3) nextline = 0;
    disp.focusLine = nextline;
  }
}

void setup() {

  Serial.begin(9600);  
  // Init LCD
  lcd.init();
  lcd.backlight();
  lcd.print("Homing...");

  // Timer 2 50KHz
  cli();
  TCCR2A = 0;
  TCCR2B = 0;
  TCNT2  = 0;
  OCR2A = 199; // 39
  TCCR2B |= (1 << WGM21);
  TCCR2B |= (0 << CS22) | (1 << CS21) | (0 << CS20);
  TIMSK2 |= (1 << OCIE2A);
  sei();

  // Init pins and home machine  
  pinMode(ENDSTOP, INPUT_PULLUP);
  pinMode(UP, INPUT_PULLUP);
  pinMode(DOWN, INPUT_PULLUP);
  pinMode(OK, INPUT_PULLUP);
  pinMode(STEP, OUTPUT);
  pinMode(DIR, OUTPUT);
  homeMachine();

  // enable the planner
  planner.init();

  // enable the display
  disp.init();
}

void loop() {
  planner.run();
  if (disp.update()) {
    updateLcd();
  }
  handleUi();
}
